# -*- coding: utf-8 -*-
"""
This contains methods for logging for GeoCube

Credits: Adopted from quest python library
"""
import logging
import os

import appdirs

_LOGGER = logging.getLogger("geocube")
_LOGGER.addHandler(logging.NullHandler())
_LOGGER.propagate = False
_LOGGER_FORMAT_STR = "%(levelname)s-%(name)s: %(message)s"


def get_logger():
    """
    Retrieve the logger for GeoCube

    Returns
    -------
    :class:`logging.Logger`
    """
    return _LOGGER


def set_log_level(level=None):
    """
    Sets the log level of the logger.

    Parameters
    ----------
    level: string, optional, default=None
        Level of logging; whichever level is chosen all higher levels
        will be logged. See: https://docs.python.org/2/library/logging.html#levels
    """
    if level is not None:
        _LOGGER.setLevel(level)
    else:
        _LOGGER.setLevel(logging.WARNING)


def _remove_log_handler(handler_type):
    for handle in _LOGGER.handlers:
        if isinstance(handle, handler_type):
            _LOGGER.removeHandler(handle)


def log_to_console(status=True, level=None):
    """Log events to the console.

    Parameters
    ----------
    status: bool, optional, default=True
        Whether logging to console should be turned on(True) or off(False)
    level: string, optional, default=None
        Level of logging; whichever level is chosen all higher levels
        will be logged. See: https://docs.python.org/2/library/logging.html#levels

    """
    set_log_level(level)
    if status:
        console_handler = logging.StreamHandler()
        formatter = logging.Formatter(_LOGGER_FORMAT_STR)
        console_handler.setFormatter(formatter)
        _LOGGER.addHandler(console_handler)
    else:
        _remove_log_handler(logging.StreamHandler)


def log_to_file(status=True, filename=None, level=None):
    """Log events to a file.

    Parameters
    ----------
    status: bool, optional, default=True
        Whether logging to console should be turned on(True) or off(False)
    filename: string, optional, default=None
        path of file to log to
    level: string, optional, default=None
        Level of logging; whichever level is chosen all higher levels
        will be logged. See: https://docs.python.org/2/library/logging.html#levels

    """
    set_log_level(level)

    if filename is None:
        filename = os.path.join(appdirs.user_log_dir("geocube", "logs"), "geocube.log")

    if status:
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError:
            pass

        file_handler = logging.FileHandler(filename)
        formatter = logging.Formatter(_LOGGER_FORMAT_STR)
        file_handler.setFormatter(formatter)
        _LOGGER.addHandler(file_handler)
    else:
        _remove_log_handler(logging.FileHandler)



# -- coding: utf-8 --
"""
This module contains tools for rasterizing vector data.
"""
import numpy
import pandas
import rasterio.features
from rasterio.enums import MergeAlg
from scipy.interpolate import Rbf, griddata
from shapely.geometry import mapping

from geocube.logger import get_logger


def _remove_missing_data(data_values, geometry_array):
    """
    Missing data causes issues with interpolation of point data
    https://github.com/corteva/geocube/issues/9

    This filters the data so those issues don't cause problems.
    """
    not_missing_data = ~pandas.isnull(data_values)
    geometry_array = geometry_array[not_missing_data]
    data_values = data_values[not_missing_data]
    return data_values, geometry_array


def rasterize_image(
    geometry_array,
    data_values,
    geobox,
    fill,
    merge_alg=MergeAlg.replace,
    filter_nan=False,
    **ignored_kwargs,
):
    """
    Rasterize a list of shapes+values for a given GeoBox.

    Parameters
    -----------
    geometry_array: geopandas.GeometryArray
        A geometry array of points.
    data_values: list
        Data values associated with the list of geojson shapes
    geobox: :obj:`datacube.utils.geometry.GeoBox`
        Transform of the resulting image.
    fill: float
        The value to fill in the grid with for nodata.
    merge_alg: `rasterio.enums.MergeAlg`, optional
        The algorithm for merging values into one cell. Default is `MergeAlg.replace`.
    filter_nan: bool, optional
        If True, will remove nodata values from the data before rasterization.
        Default is False.
    **ignored_kwargs:
        These are there to be flexible with additional rasterization methods and
        will be ignored.

    Returns
    -------
    :obj:`numpy.ndarray` or None
        The vector data in the rasterized format.

    """
    logger = get_logger()

    try:
        if filter_nan:
            data_values, geometry_array = _remove_missing_data(
                data_values, geometry_array
            )
        image = rasterio.features.rasterize(
            zip(geometry_array.apply(mapping).values, data_values),
            out_shape=(geobox.height, geobox.width),
            transform=geobox.affine,
            fill=fill,
            merge_alg=merge_alg,
            dtype=numpy.float64,
        )
        return image
    except TypeError as ter:
        if "cannot perform reduce with flexible type" in str(ter):
            logger.warning(f"{ter}")
            return None
        raise


def rasterize_points_griddata(
    geometry_array,
    data_values,
    grid_coords,
    fill,
    method="nearest",
    rescale=False,
    filter_nan=False,
    **ignored_kwargs,
):
    """
    This method uses scipy.interpolate.griddata to interpolate point data
    to a grid.

    Parameters
    ----------
    geometry_array: geopandas.GeometryArray
        A geometry array of points.
    data_values: list
        Data values associated with the list of geojson shapes
    grid_coords: dict
        Output from `rioxarray.rioxarray.affine_to_coords`
    fill: float
        The value to fill in the grid with for nodata.
    method: {‘linear’, ‘nearest’, ‘cubic’}, optional
        The method to use for interpolation in `scipy.interpolate.griddata`.
    rescale: bool, optional
        Rescale points to unit cube before performing interpolation. Default is false.
    filter_nan: bool, optional
        If True, will remove nodata values from the data before rasterization.
        Default is False.
    **ignored_kwargs:
        These are there to be flexible with additional rasterization methods and
        will be ignored.

    Returns
    -------
    :class:`numpy.ndarray`: An interpolated :class:`numpy.ndarray`.

    """
    if data_values.dtype == object:
        return None
    try:
        if filter_nan:
            data_values, geometry_array = _remove_missing_data(
                data_values, geometry_array
            )
        return griddata(
            points=(geometry_array.x, geometry_array.y),
            values=data_values,
            xi=tuple(numpy.meshgrid(grid_coords["x"], grid_coords["y"])),
            method=method,
            fill_value=fill,
            rescale=rescale,
        )
    except ValueError as ver:
        if "could not convert string to float" in str(ver):
            return None
        raise


def rasterize_points_radial(
    geometry_array,
    data_values,
    grid_coords,
    method="linear",
    filter_nan=False,
    **ignored_kwargs,
):
    """
    This method uses scipy.interpolate.Rbf to interpolate point data
    to a grid.

    Parameters
    ----------
    geometry_array: geopandas.GeometryArray
        A geometry array of points.
    data_values: list
        Data values associated with the list of geojson shapes
    grid_coords: dict
        Output from `rioxarray.rioxarray.affine_to_coords`
    method: str, optional
        The function to use for interpolation in `scipy.interpolate.Rbf`.
        {'multiquadric', 'inverse', 'gaussian', 'linear',
        'cubic', 'quintic', 'thin_plate'}
    filter_nan: bool, optional
        If True, will remove nodata values from the data before rasterization.
        Default is False.
    **ignored_kwargs:
        These are there to be flexible with additional rasterization methods and
        will be ignored.

    Returns
    -------
    :class:`numpy.ndarray`: An interpolated :class:`numpy.ndarray`.

    """
    logger = get_logger()

    try:
        if filter_nan:
            data_values, geometry_array = _remove_missing_data(
                data_values, geometry_array
            )
        interp = Rbf(geometry_array.x, geometry_array.y, data_values, function=method)
        return interp(*numpy.meshgrid(grid_coords["x"], grid_coords["y"]))
    except ValueError as ter:
        if "object arrays are not supported" in str(ter):
            logger.warning(f"{ter}")
            return None
        raise



# -- coding: utf-8 --
"""
This module is for GIS related utility functions.
"""
import json

import geopandas as gpd
import rioxarray  # noqa
from datacube.utils import geometry
from rioxarray.crs import crs_to_wkt
from shapely.geometry import box, mapping

from geocube.exceptions import VectorDataError
from geocube.logger import get_logger


def geobox_from_rio(xds):
    """This function retrieves the geobox using rioxarray extension.

    Parameters
    ----------
    xds: :obj:`xarray.DataArray` or :obj:`xarray.Dataset`
        The xarray dataset to get the geobox from.

    Returns
    -------
    :obj:`datacube.utils.geometry.GeoBox`

    """
    width, height = xds.rio.shape
    try:
        transform = xds.rio.transform()
    except AttributeError:
        transform = xds[xds.rio.vars[0]].rio.transform()
    return geometry.GeoBox(
        width=width,
        height=height,
        affine=transform,
        crs=geometry.CRS(crs_to_wkt(xds.rio.crs)),
    )


def load_vector_data(vector_data):
    """
    Parameters
    ----------
    vector_data: str or :obj:`geopandas.GeoDataFrame`
        A file path to an OGR supported source or GeoDataFrame containing
        the vector data.

    Returns
    -------
    :obj:`geopandas.GeoDataFrame` containing the vector data.

    """
    logger = get_logger()

    if isinstance(vector_data, str):
        vector_data = gpd.read_file(vector_data)
    elif not isinstance(vector_data, gpd.GeoDataFrame):
        vector_data = gpd.GeoDataFrame(vector_data)

    if vector_data.empty:
        raise VectorDataError("Empty GeoDataFrame.")
    if "geometry" not in vector_data.columns:
        raise VectorDataError(
            "'geometry' column missing. Columns in file: "
            f"{vector_data.columns.values.tolist()}"
        )

    # make sure projection is set
    if not vector_data.crs:
        vector_data.crs = "EPSG:4326"
        logger.warning(
            "Projection not defined in `vector_data`."
            " Setting to geographic (EPSG:4326)."
        )
    return vector_data


class GeoBoxMaker(object):
    """
    This class is meant for delayed GeoBox making. Stores partial information until
    all information needed is obtained.
    """

    def __init__(self, output_crs, resolution, align, geom, like):
        """Get the geobox to use for the grid.

        Parameters
        ----------
        output_crs: str, optional
            The CRS of the returned data.  If no CRS is supplied, the CRS of
             the stored data is used.
        resolution: (float,float), optional
            A tuple of the spatial resolution of the returned data.
            This includes the direction (as indicated by a positive or negative number).
            Typically when using most CRSs, the first number would be negative.
        align: (float,float), optional
            Load data such that point 'align' lies on the pixel boundary.
            Units are in the co-ordinate space of the output CRS.
            Default is (0,0)
        geom: str, optional
            A GeoJSON string for the bounding box of the data.
        like: :obj:`xarray.Dataset`, optional
            Uses the output of a previous ``load()`` to form the basis of a request for
            another product.

        """
        self.output_crs = output_crs
        self.resolution = resolution
        self.align = align
        self.geom = geom
        self.like = like

    def from_vector(self, vector_data):
        """Get the geobox to use for the grid.

        Parameters
        ----------
        vector_data: str or :obj:`geopandas.GeoDataFrame`
            A file path to an OGR supported source or GeoDataFrame
            containing the vector data.

        Returns
        -------
        :obj:`datacube.utils.geometry.GeoBox`
            The geobox for the grid to be generated from the vector data.

        """
        vector_data = load_vector_data(vector_data)

        if self.like is not None:
            assert (
                self.output_crs is None
            ), "'like' and 'output_crs' are not supported together"
            assert (
                self.resolution is None
            ), "'like' and 'resolution' are not supported together"
            assert self.align is None, "'like' and 'align' are not supported together"
            try:
                geobox = self.like.geobox
            except AttributeError:
                geobox = geobox_from_rio(self.like)
            return geobox

        if self.resolution is None:
            raise RuntimeError("Must specify 'resolution' if 'like' not specified.")

        if self.output_crs:
            crs = geometry.CRS(crs_to_wkt(self.output_crs))
        else:
            crs = geometry.CRS(crs_to_wkt(vector_data.crs))

        if self.geom is None and self.output_crs:
            geopoly = geometry.Geometry(
                mapping(
                    box(*vector_data.to_crs(crs._crs.ExportToProj4()).total_bounds)
                ),
                crs=crs,
            )
        elif self.geom is None:
            geopoly = geometry.Geometry(
                mapping(box(*vector_data.total_bounds)), crs=crs
            )

        else:
            geom_json = json.loads(self.geom)
            geom_crs = geometry.CRS(
                geom_json["crs"]["properties"]["name"]
                if "crs" in geom_json
                else "epsg:4326"
            )

            geopoly = geometry.Geometry(geom_json, crs=geom_crs)

        return geometry.GeoBox.from_geopolygon(
            geopoly, self.resolution, crs, self.align
        )



# -- coding: utf-8 --
"""
This module is an extension for xarray to provide support for vector datasets.
"""
import geopandas as gpd
import xarray
from pyproj import CRS
from shapely.wkb import dumps, loads


def from_geodataframe(in_geodataframe):
    """
    Create an xarray object from a geodataframe.
    """
    geodf = in_geodataframe.copy().set_index("geometry")
    geox = xarray.Dataset.from_dataframe(geodf)
    geox.coords["crs"] = 0
    geox.coords["crs"].attrs["crs_wkt"] = CRS.from_user_input(geodf.crs).to_wkt()
    return geox


def open_dataset(*args, **kwargs):
    """
    Open the vxarray exported netCDF file as a Dataset.

    Returns
    -------
    :obj:`xarray.Dataset`

    """
    xds = xarray.open_dataset(*args, **kwargs)
    xds.coords["geometry"] = list(map(loads, xds.coords["geometry"].values))
    return xds


class BaseVectorX:
    """This is the base for the vector GIS extension."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def _validate_operation(self):
        if set(self._obj.dims) - {"geometry"}:
            raise ValueError(
                "Extra dimensions found. Please slice/squeeze data"
                " so that the only dimension is 'geometry'."
            )

    def to_geodataframe(self):
        """
        Convert to a GeoDataFrame.

        Returns
        -------
        :obj:`geopandas.GeoDataFrame`

        """
        self._validate_operation()
        out_obj = self._obj.drop("crs")
        extra_coords = list(set(list(out_obj.coords)) - {"geometry"})
        if extra_coords:
            out_obj = out_obj.copy().reset_coords(extra_coords)
        geodf = gpd.GeoDataFrame(out_obj.to_dataframe().reset_index())
        geodf.crs = self._obj.coords["crs"].attrs["crs_wkt"]
        return geodf

    def to_netcdf(self, *args, **kwargs):
        """
        Write the data to a netCDF file.
        """
        out_obj = self._obj.copy()
        out_obj.coords["geometry"] = list(map(dumps, out_obj.coords["geometry"].values))
        out_obj.to_netcdf(*args, **kwargs)


@xarray.register_dataarray_accessor("vector")
class VectorArray(BaseVectorX):
    """This is the vector GIS extension for :class:`xarray.DataArray`"""

    def plot(self):
        self.to_geodataframe().plot(column=self._obj.name)


@xarray.register_dataset_accessor("vector")
class VectorDataset(BaseVectorX):
    """This is the vector GIS extension for :class:`xarray.Dataset`"""

    def plot(self, column):
        self._obj[column].vector.plot()



# -- coding: utf-8 --
"""
Main CLI endpoint for GeoCube
"""
import click
from click import group

import geocube.cli.commands as cmd_modules
from geocube import __version__

CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
    "token_normalize_func": lambda x: x.replace("-", "_"),
}


def check_version(ctx, _, value):
    """
    Print current version, and check for latest version.

    Called via 'geocube --version'

    :param ctx: Application context object (click.Context)
    :param value: Passed in by Click
    :return None
    """
    if not value or ctx.resilient_parsing:
        return

    click.echo(f"geocube v{__version__}")

    ctx.exit()


@group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-v",
    "--version",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=check_version,
    help="Show the current version",
)
def geocube():
    """ Top-level command and entry point into the GeoCube CLI """
    pass


def _add_subcommands():
    """
    Individual commands (and sub-commands) are encapsulated in separate files
    under /commands. Collect these command groups, and add them underneath the
    top-level command (geocube).
    """
    geocube.add_command(cmd_modules.make_geocube.make_geocube)


_add_subcommands()



# -*- coding: utf-8 -*-
"""The setup script."""

from itertools import chain

from setuptools import find_packages, setup

from geocube import __version__

with open("README.rst") as readme_file:
    readme = readme_file.read()

requirements = [
    "appdirs",
    "Click>=6.0",
    "datacube",
    "geopandas>=0.6",
    "rasterio",
    "rioxarray>=0.0.22",
    "xarray",
    "pyproj>=2",
]

test_requirements = ["pytest>=3.6", "pytest-cov", "mock"]

extras_require = {
    "dev": test_requirements
    + [
        "sphinx-click==1.1.0",
        "nbsphinx",
        "sphinx_rtd_theme",
        "black",
        "flake8",
        "pylint",
        "isort",
    ]
}
extras_require["all"] = list(chain.from_iterable(extras_require.values()))

setup(
    author="GeoCube Contributors",
    author_email="alansnow21@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Tool to convert geopandas vector data into rasterized xarray data.",
    entry_points={"console_scripts": ["geocube=geocube.cli.geocube:geocube"]},
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n",
    include_package_data=True,
    keywords="geocube",
    name="geocube",
    packages=find_packages(include=["geocube*"]),
    test_suite="test",
    tests_require=test_requirements,
    extras_require=extras_require,
    url="https://github.com/corteva/geocube",
    version=__version__,
    zip_safe=False,
    python_requires=">=3.6",
)



# -*- coding: utf-8 -*-
"""
This contains exceptions for GeoCube.
"""


class GeoCubeError(RuntimeError):
    """Base GeoCube exception class."""


class VectorDataError(GeoCubeError):
    """This is for errors in the vector data passed into GeoCube."""
